#!/usr/bin/env python

from Tkinter import *
import tkSimpleDialog, tkFileDialog, tkMessageBox
import os, sys, signal, time, getopt, shlex
import random

# UGLY GLOBAL VARIABLES...

PlayPID = None        # PID of currently playing midi
DisplayPID = None     # PID for the display program

rcFile=os.path.expanduser("~/.xpmidirc")
Version = 4.0

fullsize = 0          # command line option for fullscreen

## These are stored in the rcfile on exit. Some can
## be modified via the options menu or via ~/.xpmidirc

CurrentDir = ['.']
FavoriteDirs = []
MidiPlayer = "aplaymidi"
PlayListFile = 'gig.xpmidilst'    # a list of files to play, defaults to read midi dir
PlayOpts = "-p 20:0"      # Player options
Bcolor = "white"          # listbox colors
Fcolor = "medium blue"
killOnAbort = 1           # run "kill all notes" code on abort
displayProgram = ""
displayOptions = ""
displayDir = []

#############################################################

def usage():
    """ Display usage message and exit. """

    print "Xpmidi, GUI frontend for MIDI Player"
    print "(c) 2003-17, Bob van der Poel"
    print "Usage: xpmidi [opts] [dir | Midifiles]"
    print "Options:"
    print "   -f    start full size"
    print "   -v    display version number"
    sys.exit(0)


####################################################################
## These functions create various frames. Maintains consistency
## between different windows (and makes cleaner code??).

def makeLabelBox(parent, justify=CENTER, row=0, column=0, text=''):
    """ Create a label box. """

    f = Frame(parent)
    b = Label(f,justify=justify, text=text)
    b.grid()
    f.grid(row=row, column=column, sticky=E+W)
    f.grid_rowconfigure(0, weight=1)

    return b

def makeButtonBar(parent, row=0, column=0, buttons=(())):
    """ Create a single line frame with buttons. """

    bf=Frame(parent)
    c=0
    for txt, cmd in buttons:
        Button(bf, text=txt, height=1, command=cmd).grid(column=c, row=0, pady=5)
        c+=1
    bf.grid(row=row, column=column, sticky=W)
    return bf

def makeListBox(parent, width=50, height=20, selectmode=BROWSE, row=0, column=0):
    """ Create a list box with x and y scrollbars. """
    
    f=Frame(parent)
    ys=Scrollbar(f)
    xs=Scrollbar(f)
    lb=Listbox(f,
               bg=Bcolor,
               fg=Fcolor,
               width=width,
               height=height,
               yscrollcommand=ys.set,
               xscrollcommand=xs.set,
               exportselection=FALSE,
               selectmode=selectmode )

    ys.config(orient=VERTICAL, command=lb.yview)
    ys.grid(column=0,row=0, sticky=N+S)

    xs.config(orient=HORIZONTAL, command=lb.xview)
    xs.grid(column=1, row=1, sticky=E+W)

    lb.grid(column=1,row=0, sticky=N+E+W+S)

    f.grid(row=row, column=column, sticky=E+W+N+S) 
    f.grid_rowconfigure(0, weight=1)   
    f.grid_columnconfigure(1, weight=1)
    
    return  lb

def makeEntry(parent, label="Label", text='', column=0, row=0):

    f=Frame(parent)
    l=Label(f, anchor=E, width=15, padx=20, pady=10, text=label).grid(column=0, row=0)
    e=Entry(f, text=text, width=30)
    e.grid(column=1, row=0, sticky=E)
    e.delete(0, END)
    e.insert(END, text)
    f.grid( column=column, row=row)
    
    return e

#########################################
# We have 3 class, 1 for each window we create.


########################################
# Options dialog

class setOptions(object):

    def __init__(self):

        if PlayPID:
            return

        self.f=f=Toplevel()
        if root.winfo_viewable():  
            f.transient(root)
 
        bf=makeButtonBar(f, row=0, column=0, 
            buttons=(("Cancel", self.f.destroy), ("Apply", self.apply) ))
        self.playerEnt =  makeEntry(f, label="MIDI Player",      text=MidiPlayer,   row=1)
        self.playOptEnt = makeEntry(f, label="Player Options",   text=PlayOpts, row=2) 
        self.playerAbort = makeEntry(f, label="Kill notes on abort", text=killOnAbort, row=3)
        self.fgEnt =      makeEntry(f, label="Foreground Color", text=Fcolor,   row=4)
        self.bgEnt =      makeEntry(f, label="Background Color", text=Bcolor,   row=5)

        self.displayPrg = makeEntry(f, label="PDF Display", text=displayProgram, row=6)
        self.displayOpt = makeEntry(f, label="PDF Options", text=displayOptions, row=7)
        self.displayPath = makeEntry(f, label="PDF Path", text=', '.join(displayDir), row=8)

        f.grid_rowconfigure(1, weight=1)
        f.grid_columnconfigure(0, weight=1)

       	f.grab_set()
	root.wait_window(f)	

 
    def apply(self): 
        global MidiPlayer, PlayOpts, killOnAbort
        global Fcolor, Bcolor
        global displayProgram, displayOptions, displayDir

        MidiPlayer = self.playerEnt.get()
        PlayOpts = self.playOptEnt.get()

        killOnAbort = self.playerAbort.get()
        if killOnAbort.upper() in ("YES", "ON", "1"):
            killOnAbort = 1
        else:
            killOnAbort = 0

        displayProgram = self.displayPrg.get()
        displayOptions = self.displayOpt.get()
        fg = self.fgEnt.get()
        bg = self.bgEnt.get()

        try:
            app.lb.config(fg=fg)
            Fcolor = fg
        except TclError:
            tkMessageBox.showerror("Set Forground Color", "Illegal foreground color value")

        try:
            app.lb.config(bg=bg)
            Bcolor = bg
        except TclError:
            tkMessageBox.showerror("Set Background Color", "Illegal background color value")

        self.f.destroy()


########################################
# A listbox with the favorites directory

class selectFav(object):

    def __init__(self):

        if PlayPID:
            return

        self.f=f=Toplevel()
        if root.winfo_viewable():  
            f.transient(root)

        makeLabelBox(f, text="Select Favorite Directory", row=0, column=0)

        bf=makeButtonBar(f, row=1, column=0, 
               buttons=(("Done", self.f.destroy),
                        ("Delete", self.delete),
                        ("Add Current", self.addToFav),
                        ("Select", self.select) ) )
      
        self.lb = lb = makeListBox(f, height=10, selectmode=MULTIPLE, row=2, column=0)
        lb.bind("<Double-Button-1>", self.dclick)
        
        # Make the listbox frame expandable

        f.grid_rowconfigure(2, weight=1)
        f.grid_columnconfigure(0, weight=1)

        self.updateBox()

        f.grab_set()
        f.focus_set()
        f.wait_window(f)



    def dclick(self, w):
        """ Callback for doubleclick. Just do one dir. """

        self.doSelect( [self.lb.get(self.lb.nearest(w.y))] )


    def select(self):
        """ Callback for the 'select' button. """

        l=[]
        for n in self.lb.curselection():
            l.append(self.lb.get(int(n)))
        self.doSelect(l)


    def doSelect(self, n):
        """ Update the filelist. Called from select button or doubleclick."""

        global CurrentDir

        if n:
            CurrentDir = n
            app.updateList()
        self.f.destroy()


    def addToFav(self):
        """ Add the current directory (what's displayed) to favorites."""

        for n in CurrentDir:
            if n and not FavoriteDirs.count(n):
                FavoriteDirs.append(n)

        FavoriteDirs.sort()
        self.updateBox()

    def delete(self):
        """ Delete highlighted items, Called from 'delete' button. """

        global FavoriteDirs

        l=[]
        for n in self.lb.curselection():
            l.append(self.lb.get(int(n)))

        if l:
            if tkMessageBox.askyesno("Delete Directory",
                     "Are you sure you want to delete the "
                     "highlighted directories from the favorites list?",
                      parent=self.f):

                for n in l:
                    FavoriteDirs.remove(n)

                self.updateBox()


    def updateBox(self):
        global FavoriteDirs

        self.lb.delete(0, END)
        for n in FavoriteDirs:
            self.lb.insert(END, n)


############################
# Main display screen

class Application(object):

    def __init__(self):
        """ Create 3 frames:
               bf - the menu bar
               lf - the list box with a scroll bar
               mf - a label at the bottom with the current filename.
        """

        self.msgbox = makeLabelBox(root, justify=LEFT, row=2, column=0)

        self.lb=lb = makeListBox(root, height=28, row=1, column=0)

        self.elasped = 0
       
        bf = makeButtonBar(root, row=0, column=0, buttons=(
             ("Quit", self.quitall ),
             ("Stop", self.stopPmidi ),
             ("New Dir", self.chd),
             ("Load Playlist", self.playList),
             ("Favorites", selectFav ),
             ("Options", setOptions ) ) )

    
        self.timeButton = Button(bf, width=5, height=1)
        self.timeButton.grid(column=6, row=0, padx=10)
        
        
        # Make the listbox frame expandable

        root.grid_rowconfigure(1, weight=1)
        root.grid_columnconfigure(0, weight=1)

        # some bindings 

        lb.bind("<Return>",  self.loadfileRet)
        lb.bind("<Button-1>", self.loadfileClick)

        lb.bind('<Button-3>', self.stopPmidi)
        root.bind('<Escape>', self.stopPmidi)

        root.protocol("WM_DELETE_WINDOW", self.quitall)
        root.bind('<Control-q>', self.quitall)
        root.bind('<Alt-q>', self.quitall)

        for a in 'abcdefghijklmnopqrstuvwxyz-':
            root.bind(a, self.keyPress)

        lb.bind('<F1>', self.displayOnly)
        lb.bind('<F2>', self.rotateDisplayList)
 
        for a in "=+":
            root.bind(a, self.rndSelect)

        # end bindings

        lb.focus_force()   # make the listbox use keyboard

        self.CurrentFile = None
        self.fileList = {}    # dict of files in listbox. Key is displayed name, data=actual

        if PlayListFile:
            self.playList(PlayListFile)
        else:
            self.updateList()

        self.welcome()

    def welcome(self):
        """ Display (c) message in status box. """

        if CurrentDir:
            c=', '.join(CurrentDir)
        else:
            c = ' '
        self.msgbox.config(text="XPmidi\n%s" % c)
        ##self.timeButton.config(text="0:00")

    lastkey = ''
    lastkeytime = 0

    def rndSelect(self, ev):
        """ Callback to select a random song. """

        l = self.lb.get(0, END)

        x = random.randrange(len(l)-1)
        self.lb.select_clear(ACTIVE)   # needed to un-hilite existing selection
        self.lb.activate(x)
        self.lb.see(x)
        self.lb.select_set(x)
    

    def keyPress(self, ev):
        """ Callback for the alpha keys a..z. Finds 1st entry matching keypress. """

        c=self.lastKeyHit = ev.char.upper()

        # Timer. If there is less than 3/4 second between this key and
        # the previous we concat the keypress string. Else, start with 
        # new key.

        tm = time.time()
        delay = tm - self.lastkeytime 
        self.lastkeytime = tm     # save time of this key for next time

        if delay < .75:
            self.lastkey += c
        else:
            self.lastkey = c

        # the search target and size
        c = self.lastkey
        sz = len(c)

        l=self.lb.get(0, END)

        for x,a in enumerate(l):
            if a[0:sz].upper() >= c:
                self.lb.select_clear(ACTIVE)   # needed to un-hilite existing selection
                self.lb.activate(x)
                self.lb.see(x)
                self.lb.select_set(x)
                break
    


    """ Play a selected file. This is a listbox callback. 
        Two callback funcs are needed: one for a mouse click,
        the other for a <Return>.
    """

    def loadfileRet(self, w):
        """ Callback for <Return>. """

        self.loadfile(self.lb.get(ACTIVE) )

    def loadfileClick(self, w):
        """ Callback for <Button-1>. """

        self.lb.activate(self.lb.nearest(w.y))
        self.loadfile(self.lb.get(self.lb.nearest(w.y)))
            
    def loadfile(self, f):
        global PlayPID

        if not f:
            return

        f=self.fileList[f]
        self.stopPmidi()

        self.displayPDF(f)
        PlayPID = self.playfile(f)
        
        self.CurrentFile = f.split('/')[-1]
        cdir = '/'.join(f.split('/')[:-1])
        if cdir:
            cdir = '\nDir: ' + cdir
        else:
            cdir = '\n'
        self.msgbox.config(text="Playing: %s%s" % (self.CurrentFile, cdir))

        root.update()


    def checkfor(self):
        """ Callback for the "after" timer. Called while file is playing. """

        global PlayPID, DisplayPID

        if DisplayPID:
            try:
                os.waitpid(DisplayPID, os.WNOHANG)
            except OSError:   # our display is gone, kill the player
                DisplayPID = None
                self.stopPmidi()
                
        # PlayPID is set in loadfile(). If not set, then we are not
        # playing at all, or the "terminate midi" file is being played. 
        if PlayPID:
            t = time.time() - self.startPlayTime
            self.timeButton.config(text="%0d:%02d" % (int(t/60), int(t % 60) ))

            try:
                s = os.waitpid(PlayPID, os.WNOHANG)
                # after is one-time only so needs to reset after call
                root.after(500, self.checkfor)
                
            except OSError:  # player is gone, kill display
                if DisplayPID:
                    os.kill(DisplayPID, signal.SIGKILL) 
                DisplayPID = None
                PlayPID = None
                # reset the start time, but leave the display with
                # the time of the last play.
                self.startPlayTime = 0
                self.welcome()

    def stopPmidi(self, w=''):
        """ Stop currently playing MIDI. """

        global PlayPID, DisplayPID

        if not PlayPID and not DisplayPID:    # nothing playing, just return
            return

        if DisplayPID:
            os.kill(DisplayPID, signal.SIGKILL)
            DisplayPID = None

        cPID = PlayPID
        PlayPID = None

        self.msgbox.config(text="Stopping...%s\n" % self.CurrentFile)
        root.update_idletasks()

        """ See if last run is still running. The call to os.waitpid()
            returns a process ID and a status indication. We check the PID
            returned. If this value is equal to the current PID then
            the process has died ... and we can ignore the whole issue.
        """ 
        
        if cPID:
            try:
                pid,s = os.waitpid(cPID, os.WNOHANG)
            except OSError:
                return

            if pid:
                return
            
            # stop current player, there may be hanging notes
            x=os.kill(cPID, signal.SIGKILL)

            if killOnAbort:
                # Create a standard MIDI file which sets all notes to OFF.

                tempfile = '/tmp/xpmidi-alloff.mid'
                f=open(tempfile, 'w')

                # Standard midi file header, track count=1
                f.write( "MThd" + chr(0) + chr(0) + chr(0) + chr(6) + chr(0) + chr(1) + \
                     chr(0) + chr(1) + chr(0) +chr(0xc0) )

                # track header, len = 78 bytes
                f.write( "MTrk" + chr(0) + chr(0) + chr(0) + chr(76) )

                # all notes off for each channel
                for ch in range(0,16):
                    f.write( chr(0) + chr(0xb0 | ch) + chr(0x7b) + chr(0) )

                # midi reset sysex (at 4 ticks offset)
                f.write( chr(4) + chr(0xf0) + chr(0x05) + chr(0x7e) +
                     chr(0x7f) + chr(0x09) + chr(0x01) + chr(0xf7) )

                # EOF status event (4 tick offset)
                f.write( chr(4) + chr(0xff) + chr(0x2f) + chr(0) )
                f.close()

                self.playfile(tempfile, wait=os.P_WAIT)

                os.remove(tempfile)
                time.sleep(.5)

        self.welcome()


    def playfile(self, f, wait=os.P_NOWAIT):
        """ Call the midi player. Used by loadfile() and stopPmidi(). """

        root.after(500, self.checkfor)
        self.startPlayTime = time.time()

        # 3rd arg is a list! Player name, options and filename.
        
        ##a=os.spawnvp(os.P_NOWAIT, "sync", []) # avoid delays while playing (maybe!)
        
        op = shlex.split(PlayOpts)

        return os.spawnvp(wait, MidiPlayer, [MidiPlayer] + op + [f]  ) 

    # PDF display

    def rotateDisplayList(self, w):
        """ Callback for <F2>. Rotate display PDF list"""

        global displayDir
        if not len(displayDir):
            return

        displayDir.append(displayDir.pop(0))
        opts={'aspect':4}
        tkMessageBox.showinfo(message="DisplayPDF dir: %s" % displayDir[0])
        
    def displayOnly(self, w):
        """ Callback for <F1>. """

        self.stopPmidi()
        self.displayPDF(self.fileList[self.lb.get(ACTIVE)] )

    def displayPDF(self, midifile):
        """ Find and display a PDF for the currently playing file. """

        global DisplayPID

        if not displayProgram:
            return
        
        if DisplayPID:
            os.kill(DisplayPID, signal.SIGKILL)
            
        f = os.path.basename(midifile).replace(".mid", ".pdf")
        if len(displayDir):
            t = os.path.join(os.path.expanduser(displayDir[0]), f)
            if os.path.exists(t):
                DisplayPID = os.spawnvp(os.P_NOWAIT, displayProgram, [displayProgram] \
                        + displayOptions.split() + [t]  )
        else:
            displayPID = None


    def chd(self):
        """ Callback from <New Dir> button. Changes to new directory and updates list.  """

        global CurrentDir, PlayListFile 

        if PlayPID:
            return

        d=tkFileDialog.askdirectory(initialdir=CurrentDir)

        if d:
            PlayListFile = ''
            CurrentDir = [d]
            app.updateList()



    def quitall(self, ex=''):
        """ All done. Save current dir, stop playing and exit. """

        self.stopPmidi()

        def writeoption(s):
            print "%s = %s" % (s, eval(s))

        # The options by name and a 'type': 'list', 'string', 'integer'
        options = [
            ['FavoriteDirs',   'l'],
            ['CurrentDir',     'l'],
            ['Bcolor',         's'],
            ['Fcolor',         's'],
            ['PlayOpts',       's'],
            ['MidiPlayer',     's'],
            ['killOnAbort',    'i'],
            ['displayProgram', 's'],
            ['displayOptions', 's'],
            ['PlayListFile',    's'],
            ['displayDir',     'l']  ]

        f=open(rcFile, 'w')
        
        f.write("### XPMIDI RC FILE. Autogenerated %s, do not modify.\n\n" % time.asctime())

        for o, t in options:
            if t == 'l' or t == 'i':     # lists are just converted    
                pv = """%s""" % eval(o)
            elif t == 's':               # strings need additional love
                vv=eval(o)
                vv=vv.replace("'", "\\'")
                vv=vv.replace('"', '\\"')
                pv = """'%s'""" % vv
            f.write("%s = %s\n" % (o, pv))

        f.close()
        sys.exit()


    def playList(self, fn=''):
        global PlayListFile

        if not fn:
            infile = tkFileDialog.askopenfilename(filetypes=[("Playlists","*.xpmidilst")],
          initialdir="~")
        else:
            infile = fn
            
        try:
            inpath = open(infile, 'r')
        except:
            self.updateList()
            return

        flist=[]
        dir=''
        while 1:
            l=inpath.readline()
            if not l:
                break
            l=l.strip()
            if l.startswith("#"):
                continue

            if l.upper().startswith("DIR:"):
                dir = l[4:]
            else:
                flist.append(os.path.join(dir, l))
        
        self.updateList(flist, 0)
        

    def updateList(self, files=None, sort=1):
        """ Update the list box with with midi files in the selected directories.
    
            1. If files is NOT None, it has to be a list of midi file names. If
               there are files, skip (2).
            2. Create a list of all the files with a .mid(i) extension in all the dirs,
            3. Strip out the actual filename (less the mid ext) from each entry
               and create a dic entry with the filename as the key and the complete
               path as the data.
            4. Update the listbox 
        """

        
        if not files:
            files=[]
            for f in CurrentDir:
                for a in os.listdir(f):
                    if os.path.splitext(a)[-1].lower() in ('.mid', '.midi'):
                        files.append(os.path.join(f,a))
                        

        self.fileList = {}  # dict of filenames indexed to display name
        tlist = []          # tmp list for dislay

        # Create a display list. This is the basenames without ext.
        # of the found midi files.

        for f in files:
            a = os.path.splitext(os.path.basename(f))[0]  # base filename without ext
            self.fileList[a] = f
            tlist.append(a)
            
        self.lb.delete(0, END)

        if sort:
            tlist.sort()
       
        for f in tlist:
            self.lb.insert(END, f)
            self.lb.select_set(0)
        
        self.welcome()



###################################################
# Initial setup
####################################################


# Parse options.

try:
    opts, args = getopt.gnu_getopt(sys.argv[1:],  "vf", [])
except getopt.GetoptError:
    usage()

for o,a in opts:
    if o == '-v':
        print Version
        sys.exit(0)
    elif o == '-f':
        fullsize=1
    else:
        usage()

# Parse remaining cmd line params. This can be 1 directory name or
# a number of midi files.

fcount = 0
dcount = 0

for f in args:
    if os.path.isdir(f):
        dcount+=1
    elif os.path.isfile(f):
        fcount+=1
    else:
        print "%s is an Unknown filetype" % f
        sys.exit(1)

if dcount and fcount:
    print "You can't mix filenames and directory names on the command line."
    sys.exit(1)

if dcount > 1:
    print "Only 1 directory can be specified on the command line."
    sys.exit(1)

# Read the RC file if it exists.

if os.path.exists(rcFile):
    try:
        execfile( rcFile )
    except IOError:
        print "Error reading %s:  %s" % (rcFile, sys.exc_info()[0] )

if not CurrentDir:
    CurrentDir = ['.']

# If a dir was specified make it current

if dcount:
    CurrentDir = [os.path.abspath(os.path.expanduser(args[0]))]


# Start the tk stuff. If you want to change the font size, do it here!!!

root = Tk()
root.title("Xpmidi - pmidi frontend")
root.option_add('*font', "Georgia 20 bold")
if fullsize:
    root.geometry("%dx%s" % root.maxsize())
app=Application()

if fcount:
    app.updateList(args)

root.mainloop()




